#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "lburg.h"

static char rcsid[] = "$Id: lburg.c,v 2.2 1994/12/08 16:10:01 cwf Exp $";
static char gbuf[8192];
static char savegbuf[8192];
static char StringBuffer[8192];
static char *prefix = "";
static int Tflag = 0;
static int ntnumber = 0;
static Nonterm start = 0;
static Term terms;
static Nonterm nts;
static Rule rules;
static int nrules;
static struct block {
	struct block *link;
} *memlist;			/* list of allocated blocks */

static char *stringf(char *fmt, ...);
static void print(char *fmt, ...);
static void ckreach(Nonterm p);
static void emitclosure(Nonterm nts);
static void emitcost(Tree t, char *v);
static void emitdefs(Nonterm nts, int ntnumber);
static void emitheader(void);
static void emitkids(Rule rules, int nrules);
static void emitnts(Rule rules, int nrules);
static void emitrecalc(char *pre, Term root, Term kid);
static void emitrecord(char *pre, Rule r, int cost,int noLocalFlag);
static void emitrule(Nonterm nts);
static void emitlabel(Term terms, Nonterm start, int ntnumber);
static void emitstring(Rule rules);
static void emitstruct(Nonterm nts, int ntnumber);
static void emitterms(Term terms);
static int emittest(Tree t, char *v, char *suffix,int first);
static int buildTestString(Tree t, char *v, char *suffix);
static int findFirstTest(Rule r,Term *pterm);
static int calltest(Tree t, char *v, char *suffix,int first) ;

int main(int argc, char *argv[]) {
	int c, i;
	Nonterm p;
	
	for (i = 1; i < argc; i++)
		if (strcmp(argv[i], "-T") == 0)
			Tflag = 1;
		else if (strncmp(argv[i], "-p", 2) == 0 && argv[i][2])
			prefix = &argv[i][2];
		else if (strncmp(argv[i], "-p", 2) == 0 && i + 1 < argc)
			prefix = argv[++i];
		else if (*argv[i] == '-' && argv[i][1]) {
			yyerror("usage: %s [-T | -p prefix]... [ [ input ] output \n",
				argv[0]);
			exit(1);
		} else if (infp == NULL) {
			if (strcmp(argv[i], "-") == 0)
				infp = stdin;
			else if ((infp = fopen(argv[i], "r")) == NULL) {
				yyerror("%s: can't read `%s'\n", argv[0], argv[i]);
				exit(1);
			}
		} else if (outfp == NULL) {
			if (strcmp(argv[i], "-") == 0)
				outfp = stdout;
			if ((outfp = fopen(argv[i], "w")) == NULL) {
				yyerror("%s: can't write `%s'\n", argv[0], argv[i]);
				exit(1);
			}
		}
	if (infp == NULL)
		infp = stdin;
	if (outfp == NULL)
		outfp = stdout;
	yyparse();
	if (start)
		ckreach(start);
	for (p = nts; p; p = p->link)
		if (!p->reached)
			yyerror("can't reach nonterminal `%s'\n", p->name);
	emitheader();
	emitdefs(nts, ntnumber);
	emitstruct(nts, ntnumber);
	emitnts(rules, nrules);
	emitterms(terms);
	emitstring(rules);
	emitrule(nts);
	emitclosure(nts);
	if (start)
		emitlabel(terms, start, ntnumber);
	emitkids(rules, nrules);
	if (!feof(infp))
		while ((c = getc(infp)) != EOF)
			putc(c, outfp);
	while (memlist) {	/* for purify */
		struct block *q = memlist->link;
		free(memlist);
		memlist = q;
	}
	return errcnt > 0;
}

/* alloc - allocate nbytes or issue fatal error */
void *alloc(int nbytes) {
	struct block *p = calloc(1, sizeof *p + nbytes);

	if (p == NULL) {
		yyerror("out of memory\n");
		exit(1);
	}
	p->link = memlist;
	memlist = p;
	return p + 1;
}

/* stringf - format and save a string */
static char *stringf(char *fmt, ...) {
	va_list ap;
	char buf[512];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return strcpy(alloc(strlen(buf) + 1), buf);
}	

struct entry {
	union {
		char *name;
		struct term t;
		struct nonterm nt;
	} sym;
	struct entry *link;
} *table[211];
#define HASHSIZE (sizeof table/sizeof table[0])

/* hash - return hash number for str */
static unsigned hash(char *str) {
	unsigned h = 0;

	while (*str)
		h = (h<<1) + *str++;
	return h;
}

/* lookup - lookup symbol name */
static void *lookup(char *name) {
	struct entry *p = table[hash(name)%HASHSIZE];

	for ( ; p; p = p->link)
		if (strcmp(name, p->sym.name) == 0)
			return &p->sym;
	return 0;
}

/* install - install symbol name */
static void *install(char *name) {
	struct entry *p = alloc(sizeof *p);
	int i = hash(name)%HASHSIZE;

	p->sym.name = name;
	p->link = table[i];
	table[i] = p;
	return &p->sym;
}

/* nonterm - create a new terminal id, if necessary */
Nonterm nonterm(char *id) {
	Nonterm p = lookup(id), *q = &nts;

	if (p && p->kind == NONTERM)
		return p;
	if (p && p->kind == TERM)
		yyerror("`%s' is a terminal\n", id);
	p = install(id);
	p->kind = NONTERM;
	p->number = ++ntnumber;
	if (p->number == 1)
		start = p;
	while (*q && (*q)->number < p->number)
		q = &(*q)->link;
	assert(*q == 0 || (*q)->number != p->number);
	p->link = *q;
	*q = p;
	return p;
}

/* term - create a new terminal id with external symbol number esn */
Term term(char *id, int esn) {
	Term p = lookup(id), *q = &terms;

	if (p)
		yyerror("redefinition of terminal `%s'\n", id);
	else
		p = install(id);
	p->kind = TERM;
	p->esn = esn;
	p->arity = -1;
	while (*q && (*q)->esn < p->esn)
		q = &(*q)->link;
	if (*q && (*q)->esn == p->esn)
		yyerror("duplicate external symbol number `%s=%d'\n",
			p->name, p->esn);
	p->link = *q;
	*q = p;
	return p;
}

/* tree - create & initialize a tree node with the given fields */
Tree tree(char *id, Tree left, Tree right) {
	Tree t = alloc(sizeof *t);
	Term p = lookup(id);
	int arity = 0;

	if (left && right)
		arity = 2;
	else if (left)
		arity = 1;
	if (p == NULL && arity > 0) {
		yyerror("undefined terminal `%s'\n", id);
		p = term(id, -1);
	} else if (p == NULL && arity == 0)
		p = (Term)nonterm(id);
	else if (p && p->kind == NONTERM && arity > 0) {
		yyerror("`%s' is a nonterminal\n", id);
		p = term(id, -1);
	}
	if (p->kind == TERM && p->arity == -1)
		p->arity = arity;
	if (p->kind == TERM && arity != p->arity)
		yyerror("inconsistent arity for terminal `%s'\n", id);
	t->op = p;
	t->nterms = p->kind == TERM;
	if ((t->left = left) != NULL)
		t->nterms += left->nterms;
	if ((t->right = right) != NULL)
		t->nterms += right->nterms;
	return t;
}

/* rule - create & initialize a rule with the given fields */
Rule rule(char *id, Tree pattern, char *template, char *code) {
	Rule r = alloc(sizeof *r), *q;
	Term p = pattern->op;
	char *end;

	r->lhs = nonterm(id);
	r->packed = ++r->lhs->lhscount;
	for (q = &r->lhs->rules; *q; q = &(*q)->decode)
		;
	*q = r;
	r->pattern = pattern;
	r->ern = ++nrules;
	r->template = template;
	r->code = code;
	r->cost = strtol(code, &end, 10);
	if (*end)
		r->cost = -1;
	if (p->kind == TERM) {
		for (q = &p->rules; *q; q = &(*q)->next)
			;
		*q = r;
	} else if (pattern->left == NULL && pattern->right == NULL) {
		Nonterm p = pattern->op;
		r->chain = p->chain;
	        p->chain = r;
		if (r->cost == -1)
			yyerror("illegal nonconstant cost `%s'\n", code);
	}
	for (q = &rules; *q; q = &(*q)->link)
		;
	r->link = *q;
	*q = r;
	return r;
}

/* print - formatted output */
static void print(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	for ( ; *fmt; fmt++)
		if (*fmt == '%')
			switch (*++fmt) {
			case 'd': 
				if (outfp)
					fprintf(outfp, "%d", va_arg(ap, int)); 
				else
					sprintf(StringBuffer+strlen(StringBuffer),"%d",va_arg(ap, int)); 
				break;
			case 's':
				if (outfp)
					fputs(va_arg(ap, char *), outfp); 
				else
					sprintf(StringBuffer+strlen(StringBuffer),"%s",va_arg(ap, char *)); 
				break;
			case 'P': 
				if (outfp)
					fprintf(outfp, "%s_", prefix); 
				else
					strcat(StringBuffer,"_");
				break;
			case 'T': {
				Tree t = va_arg(ap, Tree);
				print("%S", t->op);
				if (t->left && t->right)
					print("(%T,%T)", t->left, t->right);
				else if (t->left)
					print("(%T)", t->left);
				break;
				}
			case 'R': {
				Rule r = va_arg(ap, Rule);
				print("%S: %T", r->lhs, r->pattern);
				break;
				}
			case 'S': 
				if (outfp)
					fputs(va_arg(ap, Term)->name, outfp); 
				else
					strcat(StringBuffer,va_arg(ap,Term)->name);
				break;
			case '1': case '2': case '3': case '4': case '5': {
				int n = *fmt - '0';
				while (n-- > 0) {
					if (outfp)
						putc('\t', outfp);
					else {
						char xx[2];
						xx[0] = '\t';
						xx[1] = 0;
						strcat(StringBuffer,xx);
					}
				}
				break;
				}
			default: 
				if (outfp)
					putc(*fmt, outfp); 
				else {
					char xx[2];
					xx[0] = *fmt;
					xx[1] = 0;
					strcat(StringBuffer,xx);
				}
				break;			
			}
		else {
			if (outfp)
				putc(*fmt, outfp);
			else {
				char xx[2];
				xx[0] = *fmt;
				xx[1] = 0;
				strcat(StringBuffer,xx);
			}
		}
	va_end(ap);
}

/* reach - mark all nonterminals in tree t as reachable */
static void reach(Tree t) {
	Nonterm p = t->op;

	if (p->kind == NONTERM)
		if (!p->reached)
			ckreach(p);
	if (t->left)
		reach(t->left);
	if (t->right)
		reach(t->right);
}

/* ckreach - mark all nonterminals reachable from p */
static void ckreach(Nonterm p) {
	Rule r;

        p->reached = 1;
	for (r = p->rules; r; r = r->decode)
		reach(r->pattern);
}

static int cmpRule(const void *rule1,const void *rule2)
{
	Rule r1,r2;
	FILE *saveoutfp;
	char s1[1024];

	r1 = *(Rule *)rule1;
	r2 = *(Rule *)rule2;
	saveoutfp = outfp;
	outfp = NULL;
	StringBuffer[0] = 0;
	print("%R",r1);
	s1[1023] = 0;
	strncpy(s1,StringBuffer,1023);
	StringBuffer[0] = 0;
	print("%R",r2);
	outfp = saveoutfp;
	return strcmp(s1,StringBuffer);;
}
static void printCaseArity2(Term p)
{
	Rule r,next,rvp;
	int test,ocount,i,j,k,rcount=0,needsbrace,isleft,isright;
	int thisRuleFirstTest;
	Rule *table;
	Term term;
	FILE *saveoutfp;

	for (r=p->rules; r ; r = r->next) {
		r->label = 0;
		r->tested = 0;
		r->endgroup = 0;
		if (r->pattern->nterms <= 1) {
			print("%2/* %R */\n",r);
			print("\t\tc = ");
			emitcost(r->pattern->left,  "lChild");
			emitcost(r->pattern->right, "rChild");
			print("%s;\n", r->code);
			emitrecord("\t\t", r, 0,0);
		}
		else	rcount++;
	}
	table = (Rule *)malloc((1+rcount) * sizeof(Rule *));
	i = 0;
	for (r=p->rules; r ; r = r->next) {
		if (r->pattern->nterms > 1) {
			table[i++] = r;
		}
	}
	table[rcount] = NULL;
	qsort(table,rcount,sizeof(Rule *),cmpRule);
	for (i=0; i<rcount;i++) {
		char *indent = "\t\t\0";
		isleft = isright = 0;
		r = table[i];
		if (r->label)
			print("lab%d:\n",r->label);
		if (r->endgroup)
			print("\t\t}\n");
		if (r->tested)
			print("%1");
		print("%2/* %R */\n",r);
		test = 1;
		gbuf[0] = 0;
		thisRuleFirstTest = findFirstTest(r,&term);
		next = table[i+1];
		if (next && r->tested == 0 
				&& r->pattern->nterms > 2) {
			int nextRuleFirstTest = findFirstTest(next,NULL);

			if (thisRuleFirstTest == nextRuleFirstTest) {
				rvp = next;
				j = i+1;
				needsbrace = 0;
				while (j < rcount) {
					nextRuleFirstTest = findFirstTest(table[j],NULL);
					if (nextRuleFirstTest != thisRuleFirstTest) {
						if (j != i+1) {
							table[j]->endgroup = 1;
							for (k=i; k<j;k++) {
								table[k]->tested = 1;
							}
							needsbrace = 1;
							r->tested = 1;
						}
						break;
					}
					j++;					
				}
				if (needsbrace && nextRuleFirstTest) {
					print("%2if (");
					if (thisRuleFirstTest < 0) {
						print("lChild->op");
						thisRuleFirstTest = -thisRuleFirstTest;
						isleft = 1;
					}
					else {
						print("rChild->op");
						isright = 1;
					}
					print(" == %d) { /* %S */\n",thisRuleFirstTest,term);
				}
			}
		}
		thisRuleFirstTest = findFirstTest(r,&term);
		if (r->tested) {
			isleft = (thisRuleFirstTest < 0)?1:0;
			isright = !isleft;
		}
		else print("%2");
		needsbrace = 1;
		if (r->tested && (isleft|isright)) {
			saveoutfp = outfp;
			outfp = NULL;
			StringBuffer[0] = 0;
			if (isleft)
				test = calltest(r->pattern->left,  "lChild",
					r->pattern->right->nterms ? " && " : " ",0);
			else
				test = emittest(r->pattern->left,  "lChild",
					r->pattern->right->nterms ? " && " : " ",0);
			if (isright)
				calltest(r->pattern->right, "rChild", " ",test);
			else
				emittest(r->pattern->right, "rChild", " ",test);
			outfp = saveoutfp;
			if (StringBuffer[0]) {
				print("%3if (");
				print("%s",StringBuffer);
				print("%3) {\n");
			}
			else needsbrace = 0;
		}
		else {
			print("if (");
			test = emittest(r->pattern->left,  "lChild",
				r->pattern->right->nterms ? " && " : " ",0);
			emittest(r->pattern->right, "rChild", " ",test);
			print("%2");
			print(") {\n");
		}
		indent = "\t\t\t";
		buildTestString(r->pattern->left,  "lChild",
			r->pattern->right->nterms ? " && " : " ");
		buildTestString(r->pattern->right, "rChild", " ");
		print("%sc = ", indent);
		emitcost(r->pattern->left,  "lChild");
		emitcost(r->pattern->right, "rChild");
		print("%s;\n", r->code);
		emitrecord(indent, r, 0,0);
		strcpy(savegbuf,gbuf);
		ocount=0;
		gbuf[0] = 0;
		if (savegbuf[0] && next && next->pattern->nterms > 1) {
			buildTestString(next->pattern->left,  "lChild",
				next->pattern->right->nterms ? " && " : " ");
			buildTestString(next->pattern->right, "rChild", " ");
			if (!strcmp(savegbuf,gbuf)) {
				r = next;
				i++;
				print("%3 /* %R (no test) */\n",r);
				print("%sc = ", indent);
				emitcost(r->pattern->left,  "lChild");
				emitcost(r->pattern->right, "rChild");
				print("%s;\n", r->code);
				emitrecord(indent, r, 0,0);
				gbuf[0] = 0;
			}
		}
		print("%2");
		if (r->tested)
			print("%1");
		if (needsbrace)
			print("}\n");
	}
	print("%2break;\n");
}
/* emitcase - emit one case in function state */
static void emitcase(Term p, int ntnumber) {
	Rule r;
	int test,ocount,rcount;

	print("%1case %d: /* %S */\n", p->esn, p);
	switch (p->arity) {
	case 0: case -1:
		break;
	case 1:
		print("%2%Plabel(lChild);\n");
		break;
	case 2:
		print("%2%Plabel(lChild);\n");
		print("%2%Plabel(rChild);\n");
		break;
	default: assert(0);
	}
	rcount = ocount = 0;
	if (p->arity == 2) {
		printCaseArity2(p);
		return;
	}
	for (r=p->rules; r; r = r->next) {
		char *indent = "\t\t\0";
		switch (p->arity) {
		case 0: case -1:
			if (r->code[0] != '0')
				print("%2/* %R */\n%2c = %s;\n", r, r->code);
			else
				print("%2/* %R */\n",r);
			if (r->code[0] != '0')
				emitrecord("\t\t", r, 0,0);
			else
				emitrecord("\t\t",r,0,1);
			break;
		case 1:
			print("%2 /* %R */\n",r);
			test = 1;
			if (r->pattern->nterms > 1) {
				print("%2if (");
				test = emittest(r->pattern->left, "lChild", " ",0);
				print("%2) {\n");
				indent = "\t\t\t";
			}
			if (r->pattern->nterms == 2 && r->pattern->left
			&&  r->pattern->right == NULL)
				emitrecalc(indent, r->pattern->op, r->pattern->left->op);
			print("%sc = ", indent);
			emitcost(r->pattern->left, "lChild");
			print("%s;\n", r->code);
			emitrecord(indent, r, 0,0);
			if (indent[2])
				print("%2}\n");
			break;
		case 2:
		default: assert(0);
		}
	}
	print("%2break;\n");
}

/* emitclosure - emit the closure functions */
static void emitclosure(Nonterm nts) {
	Nonterm p;

	for (p = nts; p; p = p->link)
		if (p->chain)
			print("static void %Pclosure_%S ARGS((NODEPTR_TYPE, int));\n", p);
	print("\n");
	for (p = nts; p; p = p->link)
		if (p->chain) {
			Rule r;
			print("static void %Pclosure_%S(a, c) NODEPTR_TYPE a; int c; {\n"
"%1struct %Pstate *p = STATE_LABEL(a);\n%1short *pcost = p->cost;\n", p);
			for (r = p->chain; r; r = r->chain)
				emitrecord("\t", r, r->cost,0);
			print("}\n\n");
		}
}

/* emitcost - emit cost computation for tree t */
static void emitcost(Tree t, char *v) {
	Nonterm p = t->op;

	if (p->kind == TERM) {
		if (t->left)
			emitcost(t->left,  stringf("LEFT_CHILD(%s)",  v));
		if (t->right)
			emitcost(t->right, stringf("RIGHT_CHILD(%s)", v));
	} else
		print("((struct %Pstate *)(%s->x.state))->cost[%P%S_NT] + \n\t\t\t", v, p);
}

/* emitdefs - emit nonterminal defines and data structures */
static void emitdefs(Nonterm nts, int ntnumber) {
	Nonterm p;

	for (p = nts; p; p = p->link)
		print("#define %P%S_NT %d\n", p, p->number);
	print("\n");
	print("static char *%Pntname[] = {\n%10,\n");
	for (p = nts; p; p = p->link)
		print("%1\"%S\",\n", p);
	print("%10\n};\n\n");
}

/* emitheader - emit initial definitions */
static void emitheader(void) {
	time_t timer = time(NULL);

	print("/*\ngenerated at %sby %s\n*/\n", ctime(&timer), rcsid);
	print("static void %Pkids ARGS((NODEPTR_TYPE, int, NODEPTR_TYPE[]));\n");
	print("static void %Plabel ARGS((NODEPTR_TYPE));\n");
	print("static int %Prule ARGS((void*, int));\n\n");
}

/* computekids - compute paths to kids in tree t */
static char *computekids(Tree t, char *v, char *bp, int *ip) {
	Term p = t->op;

	if (p->kind == NONTERM) {
		sprintf(bp, "\t\tkids[%d] = %s;\n", (*ip)++, v);
		bp += strlen(bp);
	} else if (p->arity > 0) {
		bp = computekids(t->left, stringf("LEFT_CHILD(%s)", v), bp, ip);
		if (p->arity == 2)
			bp = computekids(t->right, stringf("RIGHT_CHILD(%s)", v), bp, ip);
	}
	return bp;
}

/* emitkids - emit _kids */
static void emitkids(Rule rules, int nrules) {
	int i;
	Rule r, *rc = alloc((nrules + 1 + 1)*sizeof *rc);
	char **str  = alloc((nrules + 1 + 1)*sizeof *str);

	for (i = 0, r = rules; r; r = r->link) {
		int j = 0;
		char buf[1024], *bp = buf;
		*computekids(r->pattern, "p", bp, &j) = 0;
		for (j = 0; str[j] && strcmp(str[j], buf); j++)
			;
		if (str[j] == NULL)
			str[j] = strcpy(alloc(strlen(buf) + 1), buf);
		r->kids = rc[j];
		rc[j] = r;
	}
	print("static void %Pkids(p, eruleno, kids) NODEPTR_TYPE p, kids[]; int eruleno; {\n"
"%1if (!p)\n%2fatal(\"%Pkids\", \"Null tree\\n\", 0);\n"
"%1if (!kids)\n%2fatal(\"%Pkids\", \"Null kids\\n\", 0);\n"
"%1switch (eruleno) {\n");
	for (i = 0; (r = rc[i]) != NULL; i++) {
		for ( ; r; r = r->kids)
			print("%1case %d: /* %R */\n", r->ern, r);
		print("%s%2break;\n", str[i]);
	}
	print("%1default:\n%2fatal(\"%Pkids\", \"Bad rule number %%d\\n\", eruleno);\n%1}\n}\n\n");
}

/* emitlabel - emit label function */
static void emitlabel(Term terms, Nonterm start, int ntnumber) {
	int i;
	Term p;

	print("static void %Plabel(a) NODEPTR_TYPE a; {\n%1int c;\n"
"\tNODEPTR_TYPE lChild,rChild;\n"
"%1struct %Pstate *p = allocate(sizeof *p,FUNC);\n%1short *pcost = p->cost;\n\n"
"%1if (!a)\n%2fatal(\"%Plabel\", \"Null tree\\n\", 0);\n");
	print("\tlChild = LEFT_CHILD(a);\n");
	print("\trChild = RIGHT_CHILD(a);\n");
	print("%1STATE_LABEL(a) = p;\n"
"%1p->rule._stmt = 0;\n");
	for (i = 1; i <= ntnumber; i++)
		print("%1pcost[%d] =\n", i);
	print("%20x7fff;\n%1switch (OP_LABEL(a)) {\n");
	for (p = terms; p; p = p->link)
		emitcase(p, ntnumber);
	print("%1default:\n"
"%2fatal(\"%Plabel\", \"Bad terminal %%d\\n\", OP_LABEL(a));\n%1}\n}\n\n");
}

/* computents - fill in bp with _nts vector for tree t */
static char *computents(Tree t, char *bp) {
	if (t) {
		Nonterm p = t->op;
		if (p->kind == NONTERM) {
			sprintf(bp, "%s_%s_NT, ", prefix, p->name);
			bp += strlen(bp);
		} else
			bp = computents(t->right, computents(t->left,  bp));
	}
	return bp;
}

/* emitnts - emit _nts ragged array */
static void emitnts(Rule rules, int nrules) {
	Rule r;
	int i, j, *nts = alloc((nrules + 1)*sizeof *nts);
	char **str = alloc((nrules + 1)*sizeof *str);

	for (i = 0, r = rules; r; r = r->link) {
		char buf[1024];
		*computents(r->pattern, buf) = 0;
		for (j = 0; str[j] && strcmp(str[j], buf); j++)
			;
		if (str[j] == NULL) {
			print("static short %Pnts_%d[] = { %s0 };\n", j, buf);
			str[j] = strcpy(alloc(strlen(buf) + 1), buf);
		}
		nts[i++] = j;
	}
	print("\nstatic short *%Pnts[] = {\n");
	for (i = j = 0, r = rules; r; r = r->link) {
		for ( ; j < r->ern; j++)
			print("%10,%1/* %d */\n", j);
		print("%1%Pnts_%d,%1/* %d */\n", nts[i++], j++);
	}
	print("};\n\n");
}

/* emitrecalc - emit code that tests for recalculation of INDIR?(VREGP) */
static void emitrecalc(char *pre, Term root, Term kid) {
	if (root->kind == TERM && strncmp(root->name, "INDIR", 5) == 0
	&&   kid->kind == TERM &&  strcmp(kid->name,  "VREGP"   ) == 0) {
		Nonterm p;
		print("%sif (mayrecalc(a)) {\n", pre);
		print("%s%1struct %Pstate *q = a->syms[RX]->u.t.cse->x.state;\n", pre);
		/* factor in the cse q->cost and p->cost. jacob */
		print("%s%1short *qcost = q->cost;\n",pre);
		for (p = nts; p; p = p->link) {
			print("%s%1if (qcost[%P%S_NT] == 0) {\n", pre, p);
			print("%s%2pcost[%P%S_NT] = 0;\n", pre, p);
			print("%s%2p->rule.%P%S = q->rule.%P%S;\n", pre, p, p);
			print("%s%1}\n", pre);
		}
		print("%s}\n", pre);
	}
}

/* emitrecord - emit code that tests for a winning match of rule r */
static void emitrecord(char *pre, Rule r, int cost,int noLocalFlag) {
	if (Tflag)
		print("%s%Ptrace(a, %d, c + %d, pcost[%P%S_NT]);\n",
			pre, r->ern, cost, r->lhs);
	print("%sif (", pre);
	/* Avoid additions of zero. It doesn't speed up things... jacob */
	if (cost != 0)
	print("c + %d < pcost[%P%S_NT]) {\n"
"%s%1pcost[%P%S_NT] = c + %d;\n%s%1p->rule.%P%S = %d;\n",
		cost, r->lhs, pre, r->lhs, cost, pre, r->lhs,
		r->packed);
	else {
	if (noLocalFlag)
	print("pcost[%P%S_NT] > 0) {\n"
"%s%1pcost[%P%S_NT] = 0;\n%s%1p->rule.%P%S = %d;\n",
		r->lhs,pre,r->lhs,pre,r->lhs,r->packed);
	else
        print("c < pcost[%P%S_NT]) {\n"
"%s%1pcost[%P%S_NT] = c;\n%s%1p->rule.%P%S = %d;\n",
                r->lhs, pre, r->lhs, pre, r->lhs,
                r->packed);
	}
	if (r->lhs->chain) {
		if (cost)
		print("%s%1%Pclosure_%S(a, c + %d);\n", pre, r->lhs, cost);
		else {
			if (noLocalFlag)
			print("%s%1%Pclosure_%S(a, 0);\n", pre, r->lhs);
			else
			print("%s%1%Pclosure_%S(a, c);\n", pre, r->lhs);
		}
	}
	print("%s}\n", pre);
}

/* emitrule - emit decoding vectors and _rule */
static void emitrule(Nonterm nts) {
	Nonterm p;

	for (p = nts; p; p = p->link) {
		Rule r;
		print("static short %Pdecode_%S[] = {\n%10,\n", p);
		for (r = p->rules; r; r = r->decode)
			print("%1%d,\n", r->ern);
		print("};\n\n");
	}
	print("static short *DecodeTables[] = {\n");
	for (p=nts;p;p=p->link) {
		print("_decode%P%S",p);
		if (p->link)
			print(",");
		print("\n");
	}
	print("};\n");
	print("static int %Prule(state, goalnt) void *state; int goalnt; {\n"
"\tunsigned char *p;\n\tstruct %Pstate *State = (struct %Pstate *)state;\n"
"\tshort *ps;\n\n"
"%1if (goalnt < 1 || goalnt > %d)\n%2fatal(\"%Prule\", \"Bad goal nonterminal %%d\\n\", goalnt);\n"
"%1if (!state)\n%2return 0;\n", ntnumber);
	print("\tgoalnt--;\n\tp = &State->rule._stmt;\n\tps = DecodeTables[goalnt];\n");
	print("\treturn ps[p[goalnt]];\n}\n");
}

/* emitstring - emit arrays of templates, instruction flags, and rules */
static void emitstring(Rule rules) {
	Rule r;

	print("static char *%Ptemplates[] = {\n");
	print("/* 0 */%10,\n");
	for (r = rules; r; r = r->link)
		print("/* %d */%1\"%s\",%1/* %R */\n", r->ern, r->template, r);
	print("};\n");
	print("\nstatic char %Pisinstruction[] = {\n");
	print("/* 0 */%10,\n");
	for (r = rules; r; r = r->link) {
		int len = strlen(r->template);
		print("/* %d */%1%d,%1/* %s */\n", r->ern,
			len >= 2 && r->template[len-2] == '\\' && r->template[len-1] == 'n',
			r->template);
	}
	print("};\n");
	print("\nstatic char *%Pstring[] = {\n");
	print("/* 0 */%10,\n");
	for (r = rules; r; r = r->link)
		print("/* %d */%1\"%R\",\n", r->ern, r);
	print("};\n\n");
}

/* emitstruct - emit the definition of the state structure */
static void emitstruct(Nonterm nts, int ntnumber) {
	int i;
	print("struct %Pstate {\n%1short cost[%d];\n%1struct {\n", ntnumber + 1);
	for (i=0 ; nts; nts = nts->link,i++) {
		int n = 1, m = nts->lhscount;
		while ((m >>= 1) != 0)
			n++;		
			print("%2unsigned char %P%S;\n",nts);
	}
	print("%1} rule;\n};\n\n");
}

/* emitterms - emit terminal data structures */
static void emitterms(Term terms) {
	Term p;
	int k;

	print("static char %Parity[] = {\n");
	for (k = 0, p = terms; p; p = p->link) {
		for ( ; k < p->esn; k++)
			print("%10,%1/* %d */\n", k);
		print("%1%d,%1/* %d=%S */\n", p->arity < 0 ? 0 : p->arity, k++, p);
	}
	print("};\n\n");
	print("static char *%Popname[] = {\n");
	for (k = 0, p = terms; p; p = p->link) {
		for ( ; k < p->esn; k++)
			print("/* %d */%10,\n", k);
		print("/* %d */%1\"%S\",\n", k++, p);
	}
	print("};\n\n");
}

/* emittest - emit clause for testing a match */
static int emittest(Tree t, char *v, char *suffix,int first) {
	Term p = t->op;

	if (p->kind == TERM) {
		if (first == 0)
			print("%s->op == %d%s/* %S */\n", v, p->esn,
				t->nterms > 1 ? " && " : suffix, p);
		else
			print("%3%s->op == %d%s/* %S */\n", v, p->esn,
				t->nterms > 1 ? " && " : suffix, p);
		if (t->left)
			emittest(t->left, stringf("LEFT_CHILD(%s)",  v),
				t->right && t->right->nterms ? " && " : suffix,1);
		if (t->right)
			emittest(t->right, stringf("RIGHT_CHILD(%s)", v), suffix,1);
		return 1;
	}
	return 0;
}

/* buildTestString - emit clause for testing a match  in a string */
static int buildTestString(Tree t, char *v, char *suffix)
{
	Term p = t->op;

	if (p->kind == TERM) {
		sprintf(gbuf+strlen(gbuf),"%s->op == %d%s", v, p->esn,
				t->nterms > 1 ? " && " : suffix);
		if (t->left)
			buildTestString(t->left, stringf("LEFT_CHILD(%s)",  v),
				t->right && t->right->nterms ? " && " : suffix);
		if (t->right)
			buildTestString(t->right, stringf("RIGHT_CHILD(%s)", v), suffix);
		return 1;
	}
	return 0;
}

static int findFirstTest(Rule r,Term *pterm)
{
	Tree tl,tr;
	Term term1,term2;

	tl = r->pattern->left;
	tr = r->pattern->right;
	term1 = tl->op;
	term2 = tr->op;
	if (term1->kind == TERM) {
		if (pterm) *pterm = term1;
		return term1->esn * -1;
	}
	if (term2->kind == TERM) {
		if (pterm) *pterm = term2;
		return term2->esn;
	}
	return 0;
}

/* emittest - emit clause for testing a match */
static int calltest(Tree t, char *v, char *suffix,int first) 
{
	Term p = t->op;
	int r1=0,r2=0;

	if (p->kind == TERM) {
		if (t->left)
			r1 = emittest(t->left, stringf("LEFT_CHILD(%s)",  v),
				t->right && t->right->nterms ? " && " : suffix,0);
		if (t->right)
			r2 = emittest(t->right, stringf("RIGHT_CHILD(%s)", v), suffix,0);
	}
	return r1 || r2;
}
